#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <getopt.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <syscall.h>
#include <linux/kernel_stat.h>
#include <linux/kernel.h>
#include <linux/version.h>

#include <sys/wait.h>
#include <sys/param.h>
#include <sys/types.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "master.xpm"

#define SMIOC_GETKSTAT	100
#define FSHIFT		11	/* nr of bits of precision */
#define FIXED_1		(1<<FSHIFT)	/* 1.0 as fixed-point */
#define LOAD_INT(x) ((x) >> FSHIFT)
#define LOAD_FRAC(x) LOAD_INT(((x) & (FIXED_1-1)) * 100)

typedef struct {
    Display *d;
    int screen;
    Window root;
    Window win;
    Window iconwin;
    Pixmap pixmap;
    Pixmap mask;
    GC gc;
    int width;
    int height;
} Dockapp;

typedef struct {
    int his[55];
    int hisaddcnt;
    long rt_stat;
    long statlast;
    long rt_idle;
    long idlelast;
} stat_dev;

/* globals */
int fd;				/* sysmon fd for ioctl calls */
static Dockapp dockapp;
stat_dev st;
struct kernel_stat kstat;	/* where all the magic happens */
struct sysinfo si;		/* sysinfo struct for the syscall */

/* proto */
static int flush_expose(Window w);
void new_window(char *name, int width, int height);
static void redraw_window(void);
static void copy_xpm_area(int x, int y, int w, int h, int dx, int dy);

void print_usage(void);
static void get_statistics(void);
static void draw_history(int, int, int, int);

/* UI functions */
static int flush_expose(Window w)
{
    XEvent dummy;
    int i = 0;

    while (XCheckTypedWindowEvent(dockapp.d, w, Expose, &dummy))
	i++;
    return i;
}

void new_window(char *name, int width, int height)
{
    XpmAttributes attr;
    Pixel fg, bg;
    XGCValues gcval;
    XSizeHints sizehints;
    XClassHint classhint;
    XWMHints wmhints;
    XTextProperty wname;

    dockapp.width = width;
    dockapp.height = height;
    dockapp.screen = DefaultScreen(dockapp.d);
    dockapp.root = DefaultRootWindow(dockapp.d);

    sizehints.flags = USSize | USPosition;
    sizehints.x = 0;
    sizehints.y = 0;
    sizehints.width = width;
    sizehints.height = height;

    fg = BlackPixel(dockapp.d, dockapp.screen);
    bg = WhitePixel(dockapp.d, dockapp.screen);

    dockapp.win = XCreateSimpleWindow(dockapp.d, dockapp.root,
				      sizehints.x, sizehints.y,
				      sizehints.width, sizehints.height, 1,
				      fg, bg);

    dockapp.iconwin =
	XCreateSimpleWindow(dockapp.d, dockapp.win, sizehints.x,
			    sizehints.y, sizehints.width, sizehints.height,
			    1, fg, bg);

    XSetWMNormalHints(dockapp.d, dockapp.win, &sizehints);
    classhint.res_name = name;
    classhint.res_class = name;
    XSetClassHint(dockapp.d, dockapp.win, &classhint);

    XSelectInput(dockapp.d, dockapp.win,
		 ExposureMask | StructureNotifyMask);
    XSelectInput(dockapp.d, dockapp.iconwin,
		 ExposureMask | StructureNotifyMask);

    XStringListToTextProperty(&name, 1, &wname);
    XSetWMName(dockapp.d, dockapp.win, &wname);

    gcval.foreground = fg;
    gcval.background = bg;
    gcval.graphics_exposures = 0;

    dockapp.gc =
	XCreateGC(dockapp.d, dockapp.win,
		  GCForeground | GCBackground | GCGraphicsExposures,
		  &gcval);

    attr.exactColors = 0;
    attr.alloc_close_colors = 1;
    attr.closeness = 30000;
    attr.valuemask = (XpmExactColors | XpmAllocCloseColors | XpmCloseness);
    if (XpmCreatePixmapFromData(dockapp.d, dockapp.root,
				master_xpm, &dockapp.pixmap, &dockapp.mask,
				&attr) != XpmSuccess) {
	printf("can't happen!\n");
	exit(1);
    }

    XShapeCombineMask(dockapp.d, dockapp.win, ShapeBounding, 0, 0,
		      dockapp.mask, ShapeSet);
    XShapeCombineMask(dockapp.d, dockapp.iconwin, ShapeBounding, 0, 0,
		      dockapp.mask, ShapeSet);

    wmhints.initial_state = WithdrawnState;
    wmhints.icon_window = dockapp.iconwin;
    wmhints.icon_x = sizehints.x;
    wmhints.icon_y = sizehints.y;
    wmhints.window_group = dockapp.win;
    wmhints.flags =
	StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
    XSetWMHints(dockapp.d, dockapp.win, &wmhints);
    XMapWindow(dockapp.d, dockapp.win);
}

static void redraw_window(void)
{
    flush_expose(dockapp.iconwin);
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.iconwin, dockapp.gc,
	      0, 0, dockapp.width, dockapp.height, 0, 0);
    flush_expose(dockapp.win);
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.win, dockapp.gc,
	      0, 0, dockapp.width, dockapp.height, 0, 0);
}

static void copy_xpm_area(int x, int y, int w, int h, int dx, int dy)
{
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.pixmap, dockapp.gc, x, y,
	      w, h, dx, dy);
}

int main(int argc, char **argv)
{
    int ch;
    unsigned long cpubar;
    unsigned long oldcpu = 0;
    unsigned long j;
    int changed = 0;
    long starttime;
    long curtime;
    long nexttime;

    char *dispname = NULL;
    char *sysmon = NULL;

    XEvent Event;

    while ((ch = getopt(argc, argv, "d:hs:v")) != EOF) {
	switch (ch) {
	case 'd':
	    dispname = malloc(strlen(optarg) * sizeof(char));
	    strncpy(dispname, optarg, strlen(optarg));
	    break;
	case 'h':
	    print_usage();
	    exit(0);
	    break;
	case 's':
	    sysmon = malloc(strlen(optarg) * sizeof(char));
	    strncpy(sysmon, optarg, strlen(optarg));
	    break;
	case 'v':
	    printf("wmmon 1.2\n");
	    exit(0);
	    break;
	}
    }
    if (sysmon == NULL) {
	sysmon = "/dev/sysmon";
    }

    if ((fd = open(sysmon, O_RDWR)) == -1) {
	perror("cannot open sysmon device");
	fprintf(stderr, "filename was: %s\n", sysmon);
	fprintf(stderr, "make sure monitor.o module is loaded\n");
	fprintf(stderr, "hint: insmod monitor.o\n");
	exit(1);
    }

    if (!(dockapp.d = XOpenDisplay(dispname))) {
	fprintf(stderr, "Unable to open display '%s'\n", dispname);
	exit(1);
    }
    new_window("wmmon", 64, 64);
    XFlush(dockapp.d);

    /* clear history buffers */
    memset(st.his, 0, sizeof(st.his));
    st.hisaddcnt = 0;

    starttime = time(0);
    nexttime = starttime + 10;

    while (1) {

	curtime = time(0);
	get_statistics();

	/* CPU bar */
	cpubar = (st.rt_stat + st.rt_idle);
	if (cpubar != 0) {
	    cpubar = (st.rt_stat * 100) / cpubar;
	}
	cpubar = cpubar * 0.32;
	if (cpubar > 32)
	    cpubar = 32;

	if (cpubar != oldcpu) {
	    copy_xpm_area(0, 64, 32, 12, 28, 4);
	    copy_xpm_area(32, 64, cpubar, 12, 28, 4);
	    changed = 1;
	}
	oldcpu = cpubar;

	/* update history */
	if (curtime >= nexttime) {
	    nexttime += 10;

	    if (st.his[54])
		st.his[54] /= st.hisaddcnt;

	    for (j = 1; j < 55; j++) {
		st.his[j - 1] = st.his[j];
	    }
	    draw_history(54, 40, 5, 58);
	    st.his[54] = 0;
	    st.hisaddcnt = 0;
	    changed = 1;
	}

	if (changed) {
	    redraw_window();
	    changed = 0;
	}

	while (XPending(dockapp.d)) {
	    XNextEvent(dockapp.d, &Event);
	    switch (Event.type) {
	    case Expose:
		redraw_window();
		break;
	    case DestroyNotify:
		XCloseDisplay(dockapp.d);
		exit(0);
		break;
	    }
	}
	usleep(200000);
    }
    return 0;
}

static void get_statistics(void)
{
    unsigned long tmp;
    unsigned long k, istat, idle;

    ioctl(fd, SMIOC_GETKSTAT, &kstat);
    syscall(__NR_sysinfo, &si);

#ifdef WMMON_SETI
    istat = (kstat.per_cpu_user[0] + kstat.per_cpu_system[0]);
#else
    istat = (kstat.per_cpu_user[0] + kstat.per_cpu_nice[0] + kstat.per_cpu_system[0]);
#endif
    idle = (kstat.irqs[0][0] - istat);
    tmp = si.loads[0] >> (SI_LOAD_SHIFT - FSHIFT);
    k = (LOAD_INT(tmp) * 100) + LOAD_FRAC(tmp);

    st.rt_idle = idle - st.idlelast;
    st.idlelast = idle;
    st.rt_stat = istat - st.statlast;
    st.statlast = istat;
    st.his[54] += k;
    st.hisaddcnt += 1;

}

static void draw_history(int num, int size, int x_left, int y_bottom)
{

    int pixels_per_byte;
    int j, k;
    int *p;
    int d;

    pixels_per_byte = 100;
    p = st.his;
    for (j = 0; j < num; j++) {
	if (p[0] > pixels_per_byte)
	    pixels_per_byte += 100;
	p += 1;
    }

    p = st.his;

    for (k = 0; k < num; k++) {
	d = (1.0 * p[0] / pixels_per_byte) * size;

	for (j = 0; j < size; j++) {
	    if (j < d - 3)
		copy_xpm_area(2, 88, 1, 1, k + x_left, y_bottom - j);
	    else if (j < d)
		copy_xpm_area(2, 89, 1, 1, k + x_left, y_bottom - j);
	    else
		copy_xpm_area(2, 90, 1, 1, k + x_left, y_bottom - j);
	}
	p += 1;
    }

    for (j = pixels_per_byte - 100; j > 0; j -= 100) {
	for (k = 0; k < num; k++) {
	    d = (40.0 / pixels_per_byte) * j;
	    copy_xpm_area(2, 91, 1, 1, k + x_left, y_bottom - d);
	}
    }
}

void print_usage(void)
{

    printf
	("wmmon 1.0b2 modified by timecop@japan.co.jp to use monitor.o for statistics.\n"
	 "		original wmmon programming by tijno & warp\n"
	 "		memory, uptime, and i/o displays have been removed\n"
	 "		use tcsysmon to monitor memory, i/o, swap, and intr activity\n"
	 "\n" "command line switches:\n" "\n"
	 "    -d display_name		connect to display_name\n"
	 "    -h				show this help again\n"
	 "    -s device			system monitor device file\n"
	 "				( usually /dev/sysmon )\n"
	 "    -v				print version\n");
}
